"""This module contains unit tests for the SafeRandomForestClassifier."""

from __future__ import annotations

import copy
import os
import pickle

import joblib
import numpy as np
from sklearn import datasets

from aisdc.safemodel.classifiers import SafeRandomForestClassifier
from aisdc.safemodel.reporting import get_reporting_string

EXPECTED_ACC = 0.9470198675496688  # 5 estimators, min_samples_leaf 5


class DummyClassifier:
    """Dummy Classifier that always returns predictions of zero."""

    def __init__(self):
        """Empty init."""

    def fit(self, x: np.ndarray, y: np.ndarray):
        """Empty fit."""

    def predict(self, x: np.ndarray):
        """Predict all ones."""

    #     return np.ones(x.shape[0])


def get_data():
    """Returns data for testing."""
    iris = datasets.load_iris()
    x = np.asarray(iris["data"], dtype=np.float64)
    y = np.asarray(iris["target"], dtype=np.float64)
    x = np.vstack([x, (7, 2.0, 4.5, 1)])
    y = np.append(y, 4)
    return x, y


def test_randomforest_unchanged():
    """SafeRandomForestClassifier using recommended values."""
    x, y = get_data()
    model = SafeRandomForestClassifier(
        random_state=1, n_estimators=5, min_samples_leaf=5
    )
    model.fit(x, y)
    assert model.score(x, y) == EXPECTED_ACC
    msg, disclosive = model.preliminary_check()
    correct_msg = get_reporting_string(name="within_recommended_ranges")
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is False


def test_randomforest_recommended():
    """SafeRandomForestClassifier using recommended values."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.min_samples_leaf = 6
    model.fit(x, y)
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is False


def test_randomforest_unsafe_1():
    """SafeRandomForestClassifier with unsafe changes."""
    x, y = get_data()
    model = SafeRandomForestClassifier(
        random_state=1, n_estimators=5, min_samples_leaf=5
    )
    model.bootstrap = False
    model.fit(x, y)
    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter bootstrap = False identified as different than the recommended "
        "fixed value of True."
    )
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is True


def test_randomforest_unsafe_2():
    """SafeRandomForestClassifier with unsafe changes."""
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.bootstrap = True
    model.min_samples_leaf = 2
    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter min_samples_leaf = 2 identified as less than the recommended "
        "min value of 5."
    )
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is True


def test_randomforest_unsafe_3():
    """SafeRandomForestClassifier with unsafe changes."""
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    model.bootstrap = False
    model.min_samples_leaf = 2
    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter bootstrap = False identified as different than the recommended "
        "fixed value of True."
        "- parameter min_samples_leaf = 2 identified as less than the recommended "
        "min value of 5."
    )
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is True


def test_randomforest_save():
    """SafeRandomForestClassifier model saving."""
    x, y = get_data()
    model = SafeRandomForestClassifier(
        random_state=1, n_estimators=5, min_samples_leaf=5
    )
    model.fit(x, y)
    # test pickle
    model.save("rf_test.pkl")
    with open("rf_test.pkl", "rb") as file:
        pkl_model = pickle.load(file)
    assert pkl_model.score(x, y) == EXPECTED_ACC
    # test joblib
    model.save("rf_test.sav")
    with open("rf_test.sav", "rb") as file:
        sav_model = joblib.load(file)
    assert sav_model.score(x, y) == EXPECTED_ACC

    # cleanup
    for name in ("rf_test.pkl", "rf_test.sav"):
        if os.path.exists(name) and os.path.isfile(name):
            os.remove(name)


def test_randomforest_hacked_postfit():
    """SafeRandomForestClassifier changes made to parameters after fit() called."""
    x, y = get_data()
    model = SafeRandomForestClassifier(
        random_state=1, n_estimators=5, min_samples_leaf=5
    )
    model.bootstrap = False
    model.fit(x, y)
    model.bootstrap = True
    # preliminary check looks ok
    msg, disclosive = model.preliminary_check()
    correct_msg = get_reporting_string(name="within_recommended_ranges")
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is False
    # but more detailed analysis says not
    msg2, disclosive2 = model.posthoc_check()
    part1 = get_reporting_string(name="basic_params_differ", length=1)
    part2 = get_reporting_string(
        name="param_changed_from_to", key="bootstrap", val=False, cur_val=True
    )
    part3 = ""  # get_reporting_string(
    #        name="param_changed_from_to",
    #        key="base_estimator",
    #        val="DecisionTreeClassifier()",
    #        cur_val="DecisionTreeClassifier()",
    #    )
    correct_msg2 = part1 + part2 + part3
    # print(f'Correct: {correct_msg2}\n Actual: {msg2}')

    assert msg2 == correct_msg2, f"{msg2}\n should be {correct_msg2}"
    assert disclosive2 is True


def test_not_fitted():
    """Posthoc_check() called on unfitred model
    could have anything injected in classifier parameters.
    """
    unfitted_model = SafeRandomForestClassifier(random_state=1, n_estimators=5)

    # not fitted
    msg, disclosive = unfitted_model.posthoc_check()
    part1 = get_reporting_string(name="error_not_called_fit")
    part2 = get_reporting_string(name="recommend_do_not_release")
    correct_msg = part1 + part2
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive


def test_randomforest_modeltype_changed():
    """Model type has been changed after fit()
    in this this case to hide some data.
    """
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    correct_msg = ""
    # check code that tests base_estimator_
    model.fit(x, y)
    model.base_estimator = "DummyClassifier()"

    # hide some data
    for i in range(5):  # changed lengths get picked up in different test
        model.estimators_[i] = x[i, :]

    msg, disclosive = model.posthoc_check()
    # correct_msg += get_reporting_string(name="basic_params_differ",length=1)
    correct_msg = get_reporting_string(name="forest_estimators_differ", idx=5)
    correct_msg += get_reporting_string(
        name="param_changed_from_to",
        key="base_estimator",
        val="DecisionTreeClassifier()",
        cur_val="DummyClassifier()",
    )
    #    correct_msg += ("structure base_estimator has 1 differences: [('change', '', "
    #                    "(DecisionTreeClassifier(), DecisionTreeClassifier()))]"
    #                   )
    print(f"Correct: {correct_msg} Actual: {msg}")

    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive is True, "should have been flagged as disclosive"


def test_randomforest_hacked_postfit_trees_removed():
    """Tests various combinations of removing trees."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    # code that checks estimators_ : one other or both missing or different number or size
    model.fit(x, y)

    # tree removed from current
    the_estimators = model.__dict__.pop("estimators_")
    msg, disclosive = model.posthoc_check()
    correct_msg = get_reporting_string(name="current_item_removed", item="estimators_")
    # print(f'Correct: {correct_msg} Actual: {msg}')
    assert disclosive, "should be flagged as disclosive"
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"

    # trees removed from both
    _ = model.saved_model.pop("estimators_")
    msg, disclosive = model.posthoc_check()
    correct_msg = get_reporting_string(name="both_item_removed", item="estimators_")
    # print(f'Correct: {correct_msg} Actual: {msg}')
    assert disclosive, "should be flagged as disclosive"
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"

    # trees just removed from saved
    model.estimators_ = the_estimators
    msg, disclosive = model.posthoc_check()
    correct_msg = get_reporting_string(name="saved_item_removed", item="estimators_")
    # print(f'Correct: {correct_msg} Actual: {msg}')
    assert disclosive, "should be flagged as disclosive"
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"


def test_randomforest_hacked_postfit_trees_swapped():
    """Trees swapped with those from a different random forest."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    diffsizemodel = SafeRandomForestClassifier(
        random_state=1, n_estimators=5, max_depth=2
    )

    # code that checks estimators_ : one other or both missing or different number or size
    model.fit(x, y)
    diffsizemodel.fit(x, y)

    # swap saved models
    the_saved_model = copy.deepcopy(model.saved_model)
    the_saved_diffmodel = copy.deepcopy(diffsizemodel.saved_model)
    model.saved_model = copy.deepcopy(the_saved_diffmodel)
    diffsizemodel.saved_model = copy.deepcopy(the_saved_model)
    model.posthoc_check()
    msg, disclosive = diffsizemodel.posthoc_check()
    part1 = get_reporting_string(name="basic_params_differ", length=1)
    part2 = get_reporting_string(
        name="param_changed_from_to", key="max_depth", val="None", cur_val="2"
    )
    part3 = get_reporting_string(name="forest_estimators_differ", idx=5)
    part4 = ""  # get_reporting_string(
    #        name="param_changed_from_to",
    #        key="base_estimator",
    #        val="DecisionTreeClassifier()",
    #        cur_val="DecisionTreeClassifier()",
    #    )
    correct_msg = part1 + part2 + part3 + part4
    # print(f'Correct:\n{correct_msg} Actual:\n{msg}')
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive, "should be flagged as disclosive"


def test_randomforest_hacked_postfit_moretrees():
    """Trees added after fit."""
    x, y = get_data()
    model = SafeRandomForestClassifier(random_state=1, n_estimators=5)
    diffsizemodel = SafeRandomForestClassifier(random_state=1, n_estimators=10)
    model.fit(x, y)
    diffsizemodel.fit(x, y)

    # swap saved models
    the_saved_model = copy.deepcopy(model.saved_model)
    the_saved_diffmodel = copy.deepcopy(diffsizemodel.saved_model)
    model.saved_model = copy.deepcopy(the_saved_diffmodel)
    diffsizemodel.saved_model = copy.deepcopy(the_saved_model)
    msg, disclosive = diffsizemodel.posthoc_check()
    part1 = get_reporting_string(name="basic_params_differ", length=1)
    part2 = get_reporting_string(
        name="param_changed_from_to", key="n_estimators", val="5", cur_val="10"
    )
    part3 = get_reporting_string(name="different_num_estimators", num1=10, num2=5)
    part4 = ""  # get_reporting_string(
    #        name="param_changed_from_to",
    #        key="base_estimator",
    #        val="DecisionTreeClassifier()",
    #        cur_val="DecisionTreeClassifier()",
    #    )
    correct_msg = part1 + part2 + part3 + part4
    # print(f'Correct:\n{correct_msg} Actual:\n{msg}')
    assert msg == correct_msg, f"{msg}\n should be {correct_msg}"
    assert disclosive, "should be flagged as disclosive"



"""This module contains unit tests for the SafeDecisionTreeClassifier."""

from __future__ import annotations

import os
import pickle

import joblib
import numpy as np
from sklearn import datasets

from aisdc.safemodel import reporting
from aisdc.safemodel.classifiers import SafeDecisionTreeClassifier
from aisdc.safemodel.classifiers.safedecisiontreeclassifier import (
    decision_trees_are_equal,
    get_tree_k_anonymity,
)


def get_data():
    """Returns data for testing."""
    iris = datasets.load_iris()
    x = np.asarray(iris["data"], dtype=np.float64)
    y = np.asarray(iris["target"], dtype=np.float64)
    x = np.vstack([x, (7, 2.0, 4.5, 1)])
    y = np.append(y, 4)
    return x, y


def test_reporting():
    """Check that getting report name requires name."""
    correct_msg = "Error - get_reporting_string: No 'name' given"
    msg = reporting.get_reporting_string()
    assert msg == correct_msg


def test_decision_trees_are_equal():
    """Test the code that compares two decision trees."""
    x1, y = get_data()
    model1 = SafeDecisionTreeClassifier(random_state=1)
    model1.fit(x1, y)

    # same
    model2 = SafeDecisionTreeClassifier(random_state=1)
    model2.fit(x1, y)
    same, _ = decision_trees_are_equal(model1, model2)

    # one or both untrained
    model3 = SafeDecisionTreeClassifier(random_state=1, max_depth=7)
    same, msg = decision_trees_are_equal(model1, model3)
    assert same is False
    assert len(msg) > 0
    same, msg = decision_trees_are_equal(model3, model1)
    assert same is False
    assert len(msg) > 0
    same, msg = decision_trees_are_equal(model3, model3)
    assert same is True
    # assert len(msg)>0

    # different
    # x2=x1+1
    model3 = SafeDecisionTreeClassifier(random_state=1, max_depth=7)
    model3.criterion = "entropy"
    model3.fit(x1, y)
    same, msg = decision_trees_are_equal(model1, model3)
    print(f"diff msg = {msg}")
    assert same is False

    # wrong type
    same, _ = decision_trees_are_equal(model1, "aString")
    assert same is False


def test_get_tree_k_anonymity():
    """Getting k_anonymity
    50 data points randomly split in 2, single layer
    so k should be ~25.
    """
    x = np.random.rand(50, 2)
    y = np.ones(50)
    for i in range(x.shape[0]):
        if x[i][0] < 0.5:
            y[i] = 0
    model = SafeDecisionTreeClassifier(random_state=1, max_depth=1)
    model.fit(x, y)
    k = get_tree_k_anonymity(model, x)
    # print(f'k={k}')
    assert k > 10


def test_decisiontree_unchanged():
    """SafeDecisionTreeClassifier using unchanged values."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=5)
    model.fit(x, y)
    assert model.score(x, y) == 0.9668874172185431
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_safe_recommended():
    """SafeDecisionTreeClassifier using recommended values."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(
        random_state=1, max_depth=10, min_samples_leaf=10
    )
    model.min_samples_leaf = 5
    model.fit(x, y)
    assert model.score(x, y) == 0.9668874172185431
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_safe_1():
    """SafeDecisionTreeClassifier with safe changes."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(
        random_state=1, max_depth=10, min_samples_leaf=10
    )
    model.min_samples_leaf = 10
    model.fit(x, y)
    assert model.score(x, y) == 0.9536423841059603
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_safe_2():
    """SafeDecisionTreeClassifier with safe changes."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=10)
    model.fit(x, y)
    assert model.score(x, y) == 0.9536423841059603
    msg, disclosive = model.preliminary_check()
    correct_msg = "Model parameters are within recommended ranges.\n"
    assert msg == correct_msg
    assert disclosive is False


def test_decisiontree_unsafe_1():
    """SafeDecisionTreeClassifier with unsafe changes."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(
        random_state=1, max_depth=10, min_samples_leaf=10
    )
    model.min_samples_leaf = 1
    model.fit(x, y)
    assert model.score(x, y) == 1
    msg, disclosive = model.preliminary_check()
    correct_msg = (
        "WARNING: model parameters may present a disclosure risk:\n"
        "- parameter min_samples_leaf = 1 identified as less than the recommended "
        "min value of 5."
    )
    assert msg == correct_msg
    assert disclosive is True


# no longer relevant because of changed default functionality of priminary_checK()
# def test_decisiontree_unsafe_2():
#     """SafeDecisionTreeClassifier with unsafe changes - automatically fixed."""
#     x, y = get_data()
#     model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=1)
#     model.fit(x, y)
#     assert model.score(x, y) == 0.9668874172185431
#     msg, disclosive = model.preliminary_check()
#     correct_msg = "Model parameters are within recommended ranges.\n"
#     assert msg == correct_msg
#     assert disclosive is False


def test_decisiontree_save():
    """SafeDecisionTreeClassifier model saving."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=50)
    model.fit(x, y)
    assert model.score(x, y) == 0.9470198675496688
    # test pickle
    model.save("dt_test.pkl")
    with open("dt_test.pkl", "rb") as file:
        pkl_model = pickle.load(file)
    assert pkl_model.score(x, y) == 0.9470198675496688
    # test joblib
    model.save("dt_test.sav")
    with open("dt_test.sav", "rb") as file:
        sav_model = joblib.load(file)
    assert sav_model.score(x, y) == 0.9470198675496688

    # cleanup
    for name in ("dt_test.pkl", "dt_test.sav"):
        if os.path.exists(name) and os.path.isfile(name):
            os.remove(name)


def test_decisiontree_hacked_postfit():
    """SafeDecisionTreeClassifier changes made to parameters after fit() called."""
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=1)
    model.min_samples_leaf = 1
    model.fit(x, y)
    assert model.score(x, y) == 1.0
    model.min_samples_leaf = 5

    # forcing errors in internal tree checking by hacking tree
    model.tree_.max_depth = 5
    model.tree_.value[0][-1] = 1

    # At first glance the model now looks ok
    msg1, disclosive = model.preliminary_check()
    correct_msg1 = reporting.get_reporting_string(name="within_recommended_ranges")
    print(f"correct msg1: {correct_msg1}\n", f"actual  msg1: {msg1}")
    assert msg1 == correct_msg1
    assert disclosive is False

    # but on closer inspection not
    msg2, disclosive2 = model.posthoc_check()
    part1 = reporting.get_reporting_string(name="basic_params_differ", length=1)
    part2 = "parameter min_samples_leaf changed from 1 to 5 after model was fitted.\n"
    part3 = reporting.get_reporting_string(
        name="internal_attribute_differs", attr="max_depth"
    )
    part4 = reporting.get_reporting_string(
        name="internal_attribute_differs", attr="value"
    )

    correct_msg2 = part1 + part2 + part3 + part4
    print(f"Correct msg2 : {correct_msg2}\n" f"actual mesg2 : {msg2}")
    assert msg2 == correct_msg2
    assert disclosive2 is True


def test_data_hiding():
    """What if the hacking was really obscure
    like putting something in the exceptions list
     then adding data to current and saved copies.
    """
    x, y = get_data()
    model = SafeDecisionTreeClassifier(random_state=1, min_samples_leaf=5)
    model.fit(x, y)
    # now the hack
    model.examine_seperately_items = ["tree_", "bad_string"]
    model.bad_string = "something disclosive"
    model.saved_model["bad_string"] = "something disclosive"
    part1 = reporting.get_reporting_string(name="basic_params_differ", length=1)
    part2 = "('remove', 'examine_seperately_items', [(1, 'bad_string')])"
    part3 = reporting.get_reporting_string(name="unexpected_item")
    correct_msg = part1 + part2 + part3
    msg, disclosive = model.posthoc_check()
    print(f"Correct msg : {correct_msg}\n" f"actual  msg : {msg}")
    assert disclosive
    assert msg == correct_msg



"""Test the metrics."""

from __future__ import annotations

import unittest

import numpy as np
import pytest

from aisdc.metrics import (
    _div,
    _tpr_at_fpr,
    get_metrics,
    get_probabilities,
    min_max_disc,
)

# pylint: disable = invalid-name

PREDICTED_CLASS = np.array([0, 1, 0, 0, 1, 1])
TRUE_CLASS = np.array([0, 0, 0, 1, 1, 1])
PREDICTED_PROBS = np.array(
    [[0.9, 0.1], [0.4, 0.6], [0.8, 0.2], [0.55, 0.45], [0.1, 0.9], [0.01, 0.99]]
)


class DummyClassifier:
    """Mocks the predict and predict_proba methods."""

    def predict(self, _):
        """Return dummy predictions."""
        return PREDICTED_CLASS

    def predict_proba(self, _):
        """Return dummy predicted probabilities."""
        return PREDICTED_PROBS


class TestInputExceptions(unittest.TestCase):
    """Test that the metrics.py errors with a helpful error message if an
    invalid shape is supplied.
    """

    def _create_fake_test_data(self):
        y_test = np.zeros(4)
        y_test[0] = 1
        return y_test

    def test_wrong_shape(self):
        """Test the check which ensures y_pred_proba is of shape [:,:]."""
        y_test = self._create_fake_test_data()
        with pytest.raises(ValueError):
            y_pred_proba = np.zeros((4, 2, 2))
            get_metrics(y_pred_proba, y_test)

    def test_wrong_size(self):
        """Test the check which ensures y_pred_proba is of size (:,2)."""
        y_test = self._create_fake_test_data()
        with pytest.raises(ValueError):
            y_pred_proba = np.zeros((4, 4))
            get_metrics(y_pred_proba, y_test)

    def test_valid_input(self):
        """Test to make sure a valid array does not throw an exception."""
        y_test = self._create_fake_test_data()
        y_pred_proba = np.zeros((4, 2))

        returned = get_metrics(y_pred_proba, y_test)

        acc = returned["ACC"]
        auc = returned["AUC"]
        p_auc = returned["P_HIGHER_AUC"]
        tpr = returned["TPR"]

        self.assertAlmostEqual(0.75, acc)
        self.assertAlmostEqual(0.5, auc)
        self.assertAlmostEqual(0.5, p_auc)
        self.assertAlmostEqual(0.0, tpr)


class TestProbabilities(unittest.TestCase):
    """Test the checks on the input parameters of the get_probabilites function."""

    def test_permute_rows_errors(self):
        """
        Test to make sure an error is thrown when permute_rows is set to True,
        but no y_test is supplied.
        """
        clf = DummyClassifier()
        testX = []

        with pytest.raises(ValueError):
            get_probabilities(clf, testX, permute_rows=True)

    def test_permute_rows_with_permute_rows(self):
        """Test permute_rows = True succeeds."""

        clf = DummyClassifier()
        testX = np.zeros((4, 2))
        testY = np.zeros((4, 2))

        returned = get_probabilities(clf, testX, testY, permute_rows=True)

        # Check the function returns two arguments
        self.assertEqual(2, len(returned))

        # Check that the second argument is the same shape as testY
        self.assertEqual(testY.shape, returned[1].shape)

        # Check that the function is returning the right thing: predict_proba
        self.assertEqual(clf.predict_proba(testX).shape, returned[0].shape)

    def test_permute_rows_without_permute_rows(self):
        """Test permute_rows = False succeeds."""

        clf = DummyClassifier()
        testX = np.zeros((4, 2))

        y_pred_proba = get_probabilities(clf, testX, permute_rows=False)

        # Check the function returns pnly y_pred_proba
        self.assertEqual(clf.predict_proba(testX).shape, y_pred_proba.shape)


class TestMetrics(unittest.TestCase):
    """Test the metrics with some dummy predictions."""

    def test_metrics(self):
        """Test each individual metric with dummy data."""
        clf = DummyClassifier()
        testX = []
        testy = TRUE_CLASS
        y_pred_proba = get_probabilities(clf, testX, testy, permute_rows=False)
        metrics = get_metrics(y_pred_proba, testy)
        self.assertAlmostEqual(metrics["TPR"], 2 / 3)
        self.assertAlmostEqual(metrics["FPR"], 1 / 3)
        self.assertAlmostEqual(metrics["FAR"], 1 / 3)
        self.assertAlmostEqual(metrics["TNR"], 2 / 3)
        self.assertAlmostEqual(metrics["PPV"], 2 / 3)
        self.assertAlmostEqual(metrics["NPV"], 2 / 3)
        self.assertAlmostEqual(metrics["FNR"], 1 / 3)
        self.assertAlmostEqual(metrics["ACC"], 4 / 6)
        self.assertAlmostEqual(metrics["F1score"], (8 / 9) / (2 / 3 + 2 / 3))
        self.assertAlmostEqual(metrics["Advantage"], 1 / 3)
        self.assertAlmostEqual(metrics["AUC"], 8 / 9)

    def test_mia_extremecase(self):
        """Test the extreme case mia in metrics.py."""

        # create actual values
        y = np.zeros(50000)
        y[:25] = 1
        # exactly right and wrong predictions
        right = np.zeros(50000)
        right[:25] = 1
        wrong = 1 - right

        # right predictions - triggers override for very small logp
        _, _, _, pval = min_max_disc(y, right)
        self.assertEqual(-115.13, pval)

        # wrong predictions - probaility very close to 1 so logp=0
        _, _, _, pval = min_max_disc(y, wrong)
        self.assertAlmostEqual(0.0, pval)


class TestFPRatTPR(unittest.TestCase):
    """Test code that computes TPR at fixed FPR."""

    def test_tpr(self):
        """Test tpr at fpr."""
        y_true = TRUE_CLASS
        y_score = PREDICTED_PROBS[:, 1]

        tpr = _tpr_at_fpr(y_true, y_score, fpr=0)
        self.assertAlmostEqual(tpr, 2 / 3)
        tpr = _tpr_at_fpr(y_true, y_score, fpr=0.001)
        self.assertAlmostEqual(tpr, 2 / 3)
        tpr = _tpr_at_fpr(y_true, y_score, fpr=0.1)
        self.assertAlmostEqual(tpr, 2 / 3)
        tpr = _tpr_at_fpr(y_true, y_score, fpr=0.4)
        self.assertAlmostEqual(tpr, 1)
        tpr = _tpr_at_fpr(y_true, y_score, fpr=1.0)
        self.assertAlmostEqual(tpr, 1)
        tpr = _tpr_at_fpr(y_true, y_score, fpr_perc=True, fpr=100.0)
        self.assertAlmostEqual(tpr, 1)


class Test_Div(unittest.TestCase):
    """Tests the _div functionality."""

    def test_div(self):
        """Test div for y=1 and 0."""
        result = _div(8.0, 1.0, 99.0)
        self.assertAlmostEqual(result, 8.0)
        result2 = _div(8.0, 0.0, 99.0)
        self.assertAlmostEqual(result2, 99.0)


class TestExtreme(unittest.TestCase):
    """Test the extreme metrics."""

    def test_extreme_default(self):
        """Tets with the dummy data."""
        pred_probs = DummyClassifier().predict_proba(None)[:, 1]
        maxd, mind, mmd, _ = min_max_disc(TRUE_CLASS, pred_probs)

        # 10% of 6 is 1 so:
        # maxd should be 1 (the highest one is predicted as1)
        # mind should be 0 (the lowest one is not predicted as1)
        self.assertAlmostEqual(maxd, 1.0)
        self.assertAlmostEqual(mind, 0.0)
        self.assertAlmostEqual(mmd, 1.0)

    def test_extreme_higer_prop(self):
        """Tets with the dummy data but increase proportion to 0.5."""
        pred_probs = DummyClassifier().predict_proba(None)[:, 1]
        maxd, mind, mmd, _ = min_max_disc(TRUE_CLASS, pred_probs, x_prop=0.5)

        # 10% of 6 is 1 so:
        # maxd should be 1 (the highest one is predicted as1)
        # mind should be 0 (the lowest one is not predicted as1)
        self.assertAlmostEqual(maxd, 2 / 3)
        self.assertAlmostEqual(mind, 1 / 3)
        self.assertAlmostEqual(mmd, 1 / 3)



"""Test_lira_attack.py
Copyright (C) Jim Smith2022  <james.smith@uwe.ac.uk>.
"""
# pylint: disable = duplicate-code

from __future__ import annotations

import logging
import os
import shutil
import sys
from unittest import TestCase

# import json
from unittest.mock import patch

import numpy as np
from sklearn.datasets import load_breast_cancer
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import train_test_split

from aisdc.attacks import likelihood_attack
from aisdc.attacks.likelihood_attack import (  # pylint: disable = import-error
    DummyClassifier,
    LIRAAttack,
)
from aisdc.attacks.target import Target  # pylint: disable = import-error

N_SHADOW_MODELS = 20

logger = logging.getLogger(__file__)


def clean_up(name):
    """Removes unwanted files or directory."""
    if os.path.exists(name):
        if os.path.isfile(name):
            os.remove(name)
        elif os.path.isdir(name):
            shutil.rmtree(name)
        logger.info("Removed %s", name)


class TestDummyClassifier(TestCase):
    """Test the dummy classifier class."""

    @classmethod
    def setUpClass(cls):
        """Create a dummy classifier object."""
        cls.dummy = DummyClassifier()
        cls.X = np.array([[0.2, 0.8], [0.7, 0.3]])

    def test_predict_proba(self):
        """Test the predict_proba method."""
        pred_probs = self.dummy.predict_proba(self.X)
        assert np.array_equal(pred_probs, self.X)

    def test_predict(self):
        """Test the predict method."""
        expected_output = np.array([1, 0])
        pred = self.dummy.predict(self.X)
        assert np.array_equal(pred, expected_output)


class TestLiraAttack(TestCase):
    """Test the LIRA attack code."""

    @classmethod
    def setUpClass(cls):
        """Setup the common things for the class."""
        logger.info("Setting up test class")
        X, y = load_breast_cancer(return_X_y=True, as_frame=False)
        cls.train_X, cls.test_X, cls.train_y, cls.test_y = train_test_split(
            X, y, test_size=0.3
        )
        cls.target_model = RandomForestClassifier(
            n_estimators=100, min_samples_split=2, min_samples_leaf=1
        )
        cls.target_model.fit(cls.train_X, cls.train_y)
        cls.target = Target(cls.target_model)
        cls.target.add_processed_data(cls.train_X, cls.train_y, cls.test_X, cls.test_y)
        cls.target.save(path="test_lira_target")

        # Dump training and test data to csv
        np.savetxt(
            "train_data.csv",
            np.hstack((cls.train_X, cls.train_y[:, None])),
            delimiter=",",
        )
        np.savetxt(
            "test_data.csv", np.hstack((cls.test_X, cls.test_y[:, None])), delimiter=","
        )
        # dump the training and test predictions into files
        np.savetxt(
            "train_preds.csv",
            cls.target_model.predict_proba(cls.train_X),
            delimiter=",",
        )
        np.savetxt(
            "test_preds.csv", cls.target_model.predict_proba(cls.test_X), delimiter=","
        )

    def test_lira_attack(self):
        """Tests the lira code two ways."""
        attack_obj = LIRAAttack(
            n_shadow_models=N_SHADOW_MODELS,
            output_dir="test_output_lira",
            attack_config_json_file_name=os.path.join("tests", "lrconfig.json"),
        )
        attack_obj.setup_example_data()
        attack_obj.attack_from_config()
        attack_obj.example()

        attack_obj2 = LIRAAttack(
            n_shadow_models=N_SHADOW_MODELS,
            output_dir="test_output_lira",
            report_name="lira_example1_report",
        )
        attack_obj2.attack(self.target)
        output2 = attack_obj2.make_report()  # also makes .pdf and .json files
        n_shadow_models_trained = output2["attack_experiment_logger"][
            "attack_instance_logger"
        ]["instance_0"]["n_shadow_models_trained"]
        n_shadow_models = output2["metadata"]["experiment_details"]["n_shadow_models"]
        assert n_shadow_models_trained == n_shadow_models

    def test_check_and_update_dataset(self):
        """Test the code that removes items from test set with classes
        not present in training set.
        """

        attack_obj = LIRAAttack(n_shadow_models=N_SHADOW_MODELS)

        # now make test[0] have a  class not present in training set#
        local_test_y = np.copy(self.test_y)
        local_test_y[0] = 5
        local_target = Target(self.target_model)
        local_target.add_processed_data(
            self.train_X, self.train_y, self.test_X, local_test_y
        )
        unique_classes_pre = set(local_test_y)
        n_test_examples_pre = len(local_test_y)
        local_target = (
            attack_obj._check_and_update_dataset(  # pylint:disable=protected-access
                local_target
            )
        )

        unique_classes_post = set(local_target.y_test)
        n_test_examples_post = len(local_target.y_test)

        self.assertNotEqual(local_target.y_test[0], 5)
        self.assertEqual(n_test_examples_pre - n_test_examples_post, 1)
        class_diff = unique_classes_pre - unique_classes_post  # set diff
        self.assertSetEqual(class_diff, {5})

    def test_main_example(self):
        """Test command line example."""
        testargs = [
            "prog",
            "run-example",
            "--output-dir",
            "test_output_lira",
            "--report-name",
            "commandline_lira_example2_report",
        ]
        with patch.object(sys, "argv", testargs):
            likelihood_attack.main()

    def test_main_config(self):
        """Test command line with a config file."""
        testargs = [
            "prog",
            "run-attack",
            "-j",
            os.path.join("tests", "lrconfig.json"),
            "--output-dir",
            "test_output_lira",
            "--report-name",
            "commandline_lira_example1_report",
        ]
        with patch.object(sys, "argv", testargs):
            likelihood_attack.main()

    def test_main_from_configfile(self):
        """Test command line with a config file."""
        testargs = [
            "prog",
            "run-attack-from-configfile",
            "-j",
            os.path.join("tests", "lrconfig_cmd.json"),
            "-t",
            "test_lira_target",
        ]
        with patch.object(sys, "argv", testargs):
            likelihood_attack.main()

    def test_main_example_data(self):
        """Test command line example data creation."""
        testargs = [
            "prog",
            "setup-example-data",
        ]  # , "--j", os.path.join("tests","lrconfig.json")]
        with patch.object(sys, "argv", testargs):
            likelihood_attack.main()

    def test_lira_attack_failfast_example(self):
        """Tests the lira code two ways."""
        attack_obj = LIRAAttack(
            n_shadow_models=N_SHADOW_MODELS,
            output_dir="test_output_lira",
            attack_config_json_file_name=os.path.join("tests", "lrconfig.json"),
            shadow_models_fail_fast=True,
            n_shadow_rows_confidences_min=10,
        )
        attack_obj.setup_example_data()
        attack_obj.attack_from_config()
        attack_obj.example()

    def test_lira_attack_failfast_from_scratch1(self):
        """Test by training a model from scratch."""
        attack_obj = LIRAAttack(
            n_shadow_models=N_SHADOW_MODELS,
            output_dir="test_output_lira",
            report_name="lira_example2_failfast_report",
            attack_config_json_file_name=os.path.join("tests", "lrconfig.json"),
            shadow_models_fail_fast=True,
            n_shadow_rows_confidences_min=10,
        )
        attack_obj.attack(self.target)
        output = attack_obj.make_report()  # also makes .pdf and .json files
        n_shadow_models_trained = output["attack_experiment_logger"][
            "attack_instance_logger"
        ]["instance_0"]["n_shadow_models_trained"]
        n_shadow_models = output["metadata"]["experiment_details"]["n_shadow_models"]
        assert n_shadow_models_trained == n_shadow_models

    def test_lira_attack_failfast_from_scratch2(self):
        """Test by training a model from scratch."""
        attack_obj = LIRAAttack(
            n_shadow_models=150,
            output_dir="test_output_lira",
            report_name="lira_example3_failfast_report",
            attack_config_json_file_name=os.path.join("tests", "lrconfig.json"),
            shadow_models_fail_fast=True,
            n_shadow_rows_confidences_min=10,
        )
        attack_obj.attack(self.target)
        output = attack_obj.make_report()  # also makes .pdf and .json files
        n_shadow_models_trained = output["attack_experiment_logger"][
            "attack_instance_logger"
        ]["instance_0"]["n_shadow_models_trained"]
        n_shadow_models = output["metadata"]["experiment_details"]["n_shadow_models"]
        assert n_shadow_models_trained < n_shadow_models

    @classmethod
    def tearDownClass(cls):
        """Cleans up various files made during the tests."""
        names = [
            "test_output_lira",
            "output_lira",
            "outputs_lira",
            "test_lira_target",
            "test_preds.csv",
            "config.json",
            "train_preds.csv",
            "test_data.csv",
            "train_data.csv",
            "test_lira_target",
        ]
        for name in names:
            clean_up(name)
